#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>
#include <inttypes.h>

#include <string.h>
#include <stdio.h>

#include "./../curve25519_basic.c"

#define COUNT_OF(x) (sizeof(x)/sizeof(*(x)))


static void test_field_mul_avx_parallel_wrapper_with_field_mul(void **state) {
    (void) state;

    field_t a = {0};
    field_t b = {0};
    a[0] = 1;
    b[0] = 1;

    /*
    for(int i = 0; i < 10; i++){
        src[i] = (i+1) * 5;
    }
    */

    field_t expected_res = {0};
    field_mul(expected_res, a, b);

    field_t actual_res = {0};
    field_mul_avx_parallel_wrapper(actual_res, a, b);

    assert_memory_equal(expected_res, actual_res, sizeof(field_t));
}

static void test_field_square_avx_parallel_wrapper_with_field_square(void **state) {
    (void) state;

    field_t src = {0};
    src[0] = 1;

    /*
    for(int i = 0; i < 10; i++){
        src[i] = (i+1) * 5;
    }
    */

    field_t expected_res = {0};
    field_square(expected_res, src);

    field_t actual_res = {0};
    field_square_avx_parallel_wrapper(actual_res, src);

    assert_memory_equal(expected_res, actual_res, sizeof(field_t));
}

static void test_field_mul(void **state) {
    (void) state;

    byte32_t one = {
        0xa5, 0x46, 0xe3, 0x6b, 0xf0, 0x52, 0x7c, 0x9d,
        0x3b, 0x16, 0x15, 0x4b, 0x82, 0x46, 0x5e, 0xdd,
        0x62, 0x14, 0x4c, 0x0a, 0xc1, 0xfc, 0x5a, 0x18,
        0x50, 0x6a, 0x22, 0x44, 0xba, 0x44, 0x9a, 0xc4
    };

    byte32_t two = {
        0xe6, 0xdb, 0x68, 0x67, 0x58, 0x30, 0x30, 0xdb,
        0x35, 0x94, 0xc1, 0xa4, 0x24, 0xb1, 0x5f, 0x7c,
        0x72, 0x66, 0x24, 0xec, 0x26, 0xb3, 0x35, 0x3b,
        0x10, 0xa9, 0x03, 0xa6, 0xd0, 0xab, 0x1c, 0x4c
    };

    byte32_t expected = {
        0x51, 0xcf, 0x02, 0x4e, 0x5c, 0x44, 0x09, 0x91, 
        0x60, 0xf4, 0xee, 0x99, 0x12, 0x5c, 0xb0, 0x2b, 
        0xde, 0xab, 0x1e, 0xa3, 0xb4, 0x4c, 0xbd, 0x43, 
        0x90, 0x28, 0x12, 0x4b, 0x58, 0x0f, 0x15, 0x10
    };

    field_t field_one = {0};
    field_t field_two = {0};

    decode_point(field_one, one);
    decode_point(field_two, two);

    field_t field_actual = {0};
    field_mul(field_actual, field_one, field_two);

    byte32_t actual = {0};
    encode_point(actual, field_actual);
    
    assert_memory_equal(actual, expected, sizeof(byte32_t));
}

static void print_byte32_t(byte32_t value)
{
    for (int i = 0; i < 32; i++) {
        if (i % 8 == 0) {
            printf("\n");
        }
        printf("0x%02x", value[i]);
        if (i < 31) {
            printf(", ");
        }
    }
    printf("\n");
}

static void rfc_vector_1(void **state) {
    (void) state;

    uint8_t n[32] = {
        0xa5, 0x46, 0xe3, 0x6b, 0xf0, 0x52, 0x7c, 0x9d,
        0x3b, 0x16, 0x15, 0x4b, 0x82, 0x46, 0x5e, 0xdd,
        0x62, 0x14, 0x4c, 0x0a, 0xc1, 0xfc, 0x5a, 0x18,
        0x50, 0x6a, 0x22, 0x44, 0xba, 0x44, 0x9a, 0xc4
    };

    uint8_t base_point[32] = {
        0xe6, 0xdb, 0x68, 0x67, 0x58, 0x30, 0x30, 0xdb,
        0x35, 0x94, 0xc1, 0xa4, 0x24, 0xb1, 0x5f, 0x7c,
        0x72, 0x66, 0x24, 0xec, 0x26, 0xb3, 0x35, 0x3b,
        0x10, 0xa9, 0x03, 0xa6, 0xd0, 0xab, 0x1c, 0x4c
    };

    uint8_t expected[32] = {
        0xc3, 0xda, 0x55, 0x37, 0x9d, 0xe9, 0xc6, 0x90,
        0x8e, 0x94, 0xea, 0x4d, 0xf2, 0x8d, 0x08, 0x4f,
        0x32, 0xec, 0xcf, 0x03, 0x49, 0x1c, 0x71, 0xf7,
        0x54, 0xb4, 0x07, 0x55, 0x77, 0xa2, 0x85, 0x52
    };

    uint8_t actual[32];
    crypto_scalarmult(actual, n, base_point);
    assert_memory_equal(actual, expected, sizeof(uint8_t) * 32);
}

static void rfc_vector_2(void **state) {
    (void) state;

    uint8_t n[32] = {
        0x4b, 0x66, 0xe9, 0xd4, 0xd1, 0xb4, 0x67, 0x3c,
        0x5a, 0xd2, 0x26, 0x91, 0x95, 0x7d, 0x6a, 0xf5,
        0xc1, 0x1b, 0x64, 0x21, 0xe0, 0xea, 0x01, 0xd4,
        0x2c, 0xa4, 0x16, 0x9e, 0x79, 0x18, 0xba, 0x0d
    };

    uint8_t base_point[32] = {
        0xe5, 0x21, 0x0f, 0x12, 0x78, 0x68, 0x11, 0xd3,
        0xf4, 0xb7, 0x95, 0x9d, 0x05, 0x38, 0xae, 0x2c,
        0x31, 0xdb, 0xe7, 0x10, 0x6f, 0xc0, 0x3c, 0x3e,
        0xfc, 0x4c, 0xd5, 0x49, 0xc7, 0x15, 0xa4, 0x93
    };

    uint8_t expected[32] = {
        0x95, 0xcb, 0xde, 0x94, 0x76, 0xe8, 0x90, 0x7d,
        0x7a, 0xad, 0xe4, 0x5c, 0xb4, 0xb8, 0x73, 0xf8,
        0x8b, 0x59, 0x5a, 0x68, 0x79, 0x9f, 0xa1, 0x52,
        0xe6, 0xf8, 0xf7, 0x64, 0x7a, 0xac, 0x79, 0x57
    };

    uint8_t actual[32];
    crypto_scalarmult(actual, n, base_point);
    assert_memory_equal(actual, expected, sizeof(uint8_t) * 32);
}

static void rfc_vector_iterate(void **state) {
    (void) state;


    uint8_t n[32] = {
        0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };

    uint8_t base_point[32] = {
        0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };

    uint8_t expected_1[32] = {
        0x42, 0x2c, 0x8e, 0x7a, 0x62, 0x27, 0xd7, 0xbc,
        0xa1, 0x35, 0x0b, 0x3e, 0x2b, 0xb7, 0x27, 0x9f,
        0x78, 0x97, 0xb8, 0x7b, 0xb6, 0x85, 0x4b, 0x78,
        0x3c, 0x60, 0xe8, 0x03, 0x11, 0xae, 0x30, 0x79
    };

    uint8_t expected_1000[32] = {
        0x68, 0x4c, 0xf5, 0x9b, 0xa8, 0x33, 0x09, 0x55,
        0x28, 0x00, 0xef, 0x56, 0x6f, 0x2f, 0x4d, 0x3c,
        0x1c, 0x38, 0x87, 0xc4, 0x93, 0x60, 0xe3, 0x87,
        0x5f, 0x2e, 0xb9, 0x4d, 0x99, 0x53, 0x2c, 0x51
    };

    uint8_t actual[32];

    for (int i = 1; i <= 1000; i++) {
        crypto_scalarmult(actual, n, base_point);
        memcpy(base_point, n, sizeof(uint8_t) * 32);
        memcpy(n, actual, sizeof(uint8_t) * 32);

        if (i == 1) {
            assert_memory_equal(actual, expected_1, sizeof(uint8_t) * 32);
        }

        if (i == 1000) {
            assert_memory_equal(actual, expected_1000, sizeof(uint8_t) * 32);
        }
    }
}

int main() {
    const struct CMUnitTest tests[] = {
        cmocka_unit_test(test_field_mul_avx_parallel_wrapper_with_field_mul),
        cmocka_unit_test(test_field_square_avx_parallel_wrapper_with_field_square),
        cmocka_unit_test(test_field_mul),
        cmocka_unit_test(rfc_vector_1),
        cmocka_unit_test(rfc_vector_2),
        cmocka_unit_test(rfc_vector_iterate)
    };
    return cmocka_run_group_tests(tests, NULL, NULL);
}
